package sep.io;

import java.io.*;

/**
 * This class stores and retrieves the data values that underlie a
 * regular data set.
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * Bugs: <P>
 * All methods should probably be protected (only accessible through
 *     RegularData.<P>
 * The data should be able to hold everything, not only floats.<P>
 *
 * @author Matthias Schwab
 */
public class DataKeeper {
  static boolean local = false;

  Object data;
  int    length;
  String type;

  /**
   * the logic here will probably be CHANGED when we install
   * JDK1.1, since it has more support for dealing with native methods.
   * <P>
   * THIS SHOULD PROBABLY BE A RELATIVE PATH to LD_LIBRARY_PATH,
   * but that does not work, since System.load() requires a full
   * path name.  The alternative is to use System.loadLibrary() with
   * LD_LIBRARY_PATH set to the directory containing the .so file.
   * Unfortunately, I have not been able to get this to work.
   */
  static {
    try {
      // If this throws an exception, we are probably non-local
      System.getProperty("java.home");
      // The following (commented out) seems to ALWAYS throw an exception
      //      System.getSecurityManager().checkPropertyAccess("java.home");
      String OpSys   = System.getProperty("os.name");
      // if local and Linux, load in a native routine to speed up write()
      if (OpSys.equals("Linux")) {
	local = true;
	//String libDir  = "/homes/sep/matt/jest/sep/io/";
	String libDir  = "/homes/sep/matt/jest/sep/";
	String libFile = "DataKeeper.so";
	System.load(libDir + OpSys + "/" + libFile);
      }
    }
    catch (Exception e) {
      System.err.println(e);
      System.err.println("Using non-native code for data writing.");
    }
    /*
    try {
      System.getSecurityManager().checkLink(libName);
      System.load(libName);
    }
    catch (Exception e) { System.err.println(e); }
    */
  }

  public DataKeeper() {};

  public DataKeeper( Object data, int length, String type) {
    this.data   = data;
    this.length = length;
    this.type   = type;
  }

  /**
   * Ask Joel if he knows how to do the runtime assignment of an array
   * to an object of type Object
   */
  public void read(DataInputStream dinstr, int length, String type) 
       throws IOException {

    this.length = length;
    this.type   = type;

    if (type.equals("float")) {
      float[] dummy = new float[length];
      for (int i = 0; i < length; i++) {
	dummy[i] = dinstr.readFloat();
      }
      this.data = dummy;
    } 
    else {
      System.out.println("DataKeeper: other data types not implemented yet");
    }
  }

  /**
   * I want this to be able to write and read only parts of a stream
   * (much like sreed). I also would like to make sure that data type
   * etc are consistent with header information.
   */

  native void floats2bytes(float[] f, byte[] b, int len);

  public void write(DataOutputStream dout) throws IOException {
    float[] dat = (float[]) data;
    byte[] bytes = new byte[4 * dat.length];
    //System.err.println("I am in DataKeeper.write()");
    if (local) {
      floats2bytes(dat, bytes, bytes.length);
      dout.write(bytes);
    }
    else {
      for (int i = 0; i < dat.length; i++)
	dout.writeFloat(dat[i]);
    }
    // this is needed if DataStream is buffered
    dout.flush();
  }

  public String toString() { 
    return ("Data:  " + data.toString() + "\n" + 
	    "Length:" + length + "\n" + 
	    "Type:  " + type
	   );
  }
  public boolean equals(DataKeeper para) {
    System.out.println("DataKeeper: equals not implemented yet");
    return false ;
  }

  public Object getData() {
    return data;
  }
  public void setData(float[] input) { // this should be for all types 
    if (length == input.length) {
      for (int i=0; i < input.length ; i++) {
	data = input;
      }
    }
  }

  public int getLength() {
    return length;
  }
  public String getType() {
    return type;
  }
}





