package sep.io;

import java.io.*;

/**
 * This class stores and retrieves the data values that underlie a
 * regular data set.
 * <BR><!--#include file="./source.inc"--><BR>
 *
 * Bugs: <P>
 * All methods should probably be protected (only accessible through
 *     RegularData.<P>
 * The data should be able to hold everything, not only floats.<P>
 *
 * @author Matthias Schwab
 */
public class DataKeeperJ {
  static boolean local = false;
  int            dim   = 0; 

  float[]           data1 = null; 
  float[][]         data2 = null; 
  float[][][]       data3 = null; 
  float[][][][]     data4 = null; 
  float[][][][][]   data5 = null; 
  float[][][][][][] data6 = null; 

  /**
   * the logic here will probably be CHANGED when we install
   * JDK1.1, since it has more support for dealing with native methods.
   * <P>
   * THIS SHOULD PROBABLY BE A RELATIVE PATH to LD_LIBRARY_PATH,
   * but that does not work, since System.load() requires a full
   * path name.  The alternative is to use System.loadLibrary() with
   * LD_LIBRARY_PATH set to the directory containing the .so file.
   * Unfortunately, I have not been able to get this to work.
   */
  static {
    try {
      // If this throws an exception, we are probably non-local
      System.getProperty("java.home");
      // The following (commented out) seems to ALWAYS throw an exception
      //      System.getSecurityManager().checkPropertyAccess("java.home");
      String OpSys   = System.getProperty("os.name");
      // if local and Linux, load in a native routine to speed up write()
      if (OpSys.equals("Linux")) {
	local = true;
	//String libDir  = "/homes/sep/matt/jest/sep/io/";
	String libDir  = "/homes/sep/matt/jest/sep/";
	String libFile = "DataKeeperJ.so";
	System.load(libDir + OpSys + "/" + libFile);
      }
    }
    catch (Exception e) {
      System.err.println(e);
      System.err.println("Using non-native code for data writing.");
    }
    /*
    try {
      System.getSecurityManager().checkLink(libName);
      System.load(libName);
    }
    catch (Exception e) { System.err.println(e); }
    */
  }

  /** to read prepare the DataKeeperJ to read in data */
  public DataKeeperJ(int[] n) {
    dim = n.length;
    switch (n.length) {
    case 1: data1 = new float[n[0]]                              ; break; 
    case 2: data2 = new float[n[0]][n[1]]                        ; break; 
    case 3: data3 = new float[n[0]][n[1]][n[2]]                  ; break; 
    case 4: data4 = new float[n[0]][n[1]][n[2]][n[3]]            ; break; 
    case 5: data5 = new float[n[0]][n[1]][n[2]][n[3]][n[4]]      ; break; 
    case 6: data6 = new float[n[0]][n[1]][n[2]][n[3]][n[4]][n[5]]; break; 
    default: System.err.println("Error: dimension not implemented yet");
    }
  }

  /** to write out usually (not sure if needed) */
  public DataKeeperJ(float[]           data) {
    this.data1 = data;
    this.dim   = 1;
  }
  public DataKeeperJ(float[][]         data) {
    this.data2 = data;
    this.dim   = 2;
  }
  public DataKeeperJ(float[][][]       data) {
    this.data3 = data;
    this.dim   = 3;
  }
  public DataKeeperJ(float[][][][]     data) {
    this.data4 = data;
    this.dim   = 4;
  }
  public DataKeeperJ(float[][][][][]   data) {
    this.data5 = data;
    this.dim   = 5;
  }
  public DataKeeperJ(float[][][][][][] data) {
    this.data6 = data;
    this.dim   = 6;
  }

  /**
   * Ask Joel if he knows how to do the runtime assignment of an array
   * to an object of type Object
   */
  public void read(DataInputStream dinstr) throws IOException {
    switch (dim) {
    case 1: 
      for (int i0 = 0; i0 < data1.length; i0++) 
	data1[i0]  = dinstr.readFloat();
      break;
    case 2: 
      for (int i0 = 0; i0 < data2   .length; i0++) 
      for (int i1 = 0; i1 < data2[0].length; i1++) 
	data2[i0][i1] = dinstr.readFloat();
      break;
    case 3: 
      for (int i0 = 0; i0 < data3               .length; i0++) 
      for (int i1 = 0; i1 < data3[0]            .length; i1++) 
      for (int i2 = 0; i2 < data3[0][0]         .length; i2++) 
	data3[i0][i1][i2] = dinstr.readFloat();
      break;
    case 4: 
      for (int i0 = 0; i0 < data4               .length; i0++) 
      for (int i1 = 0; i1 < data4[0]            .length; i1++) 
      for (int i2 = 0; i2 < data4[0][0]         .length; i2++) 
      for (int i3 = 0; i3 < data4[0][0][0]      .length; i3++) 
	data4[i0][i1][i2][i3] = dinstr.readFloat();
      break;
    case 5: 
      for (int i0 = 0; i0 < data5               .length; i0++) 
      for (int i1 = 0; i1 < data5[0]            .length; i1++) 
      for (int i2 = 0; i2 < data5[0][0]         .length; i2++) 
      for (int i3 = 0; i3 < data5[0][0][0]      .length; i3++) 
      for (int i4 = 0; i4 < data5[0][0][0][0]   .length; i4++) 
	data5[i0][i1][i2][i3][i4] = dinstr.readFloat();
      break;
    case 6: 
      for (int i0 = 0; i0 < data6               .length; i0++) 
      for (int i1 = 0; i1 < data6[0]            .length; i1++) 
      for (int i2 = 0; i2 < data6[0][0]         .length; i2++) 
      for (int i3 = 0; i3 < data6[0][0][0]      .length; i3++) 
      for (int i4 = 0; i4 < data6[0][0][0][0]   .length; i4++) 
      for (int i5 = 0; i5 < data6[0][0][0][0][0].length; i5++) 
	data6[i0][i1][i2][i3][i4][i5] = dinstr.readFloat();
      break;
    default: System.err.println("Error: dimension not implemented yet");
    }
  }

  /**
   * I want this to be able to write and read only parts of a stream
   * (much like sreed). I also would like to make sure that data type
   * etc are consistent with header information.
   */

  native void floats2bytes(float[] f, byte[] b, int len);

  public void write(float[] data, DataOutputStream dout) throws IOException {
    byte[] bytes = new byte[4 * data.length];
    //System.err.println("I am in DataKeeperJ.write()");
    if (local) {
      floats2bytes(data, bytes, bytes.length);
      dout.write(bytes);
    }
    else {
      for (int i = 0; i < data.length; i++)
	dout.writeFloat(data[i]);
    }
  }

  public void write(DataOutputStream dout) throws IOException {
    switch (dim) {
    case 1: 
        write(data1, dout); 
      break;
    case 2: 
      for (int i0 = 0; i0 < data2   .length; i0++) 
	write(data2[i0], dout); 
      break;
    case 3: 
      for (int i0 = 0; i0 < data3               .length; i0++) 
      for (int i1 = 0; i1 < data3[0]            .length; i1++) 
	write(data3[i0][i1], dout); 
      break;
    case 4: 
      for (int i0 = 0; i0 < data4               .length; i0++) 
      for (int i1 = 0; i1 < data4[0]            .length; i1++) 
      for (int i2 = 0; i2 < data4[0][0]         .length; i2++) 
	write(data4[i0][i1][i2], dout); 
      break;
    case 5: 
      for (int i0 = 0; i0 < data5               .length; i0++) 
      for (int i1 = 0; i1 < data5[0]            .length; i1++) 
      for (int i2 = 0; i2 < data5[0][0]         .length; i2++) 
      for (int i3 = 0; i3 < data5[0][0][0]      .length; i3++) 
	write(data5[i0][i1][i2][i3], dout);
      break;
    case 6: 
      for (int i0 = 0; i0 < data6               .length; i0++) 
      for (int i1 = 0; i1 < data6[0]            .length; i1++) 
      for (int i2 = 0; i2 < data6[0][0]         .length; i2++) 
      for (int i3 = 0; i3 < data6[0][0][0]      .length; i3++) 
      for (int i4 = 0; i4 < data6[0][0][0][0]   .length; i4++) 
	write(data6[i0][i1][i2][i3][i4], dout);
      break;
    default: System.err.println("Error: dimension not implemented yet");
    }
    // this is needed if DataStream is buffered
    dout.flush();
  }

  public String toString() { 
    StringBuffer buf = new StringBuffer(); 
    buf.append("Dimension: "+ dim);
    buf.append("Data : ");
    switch(dim) {
    case 1: buf.append(data1); break;
    case 2: buf.append(data2); break;
    case 3: buf.append(data3); break;
    case 4: buf.append(data4); break;
    case 5: buf.append(data5); break;
    case 6: buf.append(data6); break;
    default: System.err.println("Error: dim not implemented yet"); break;
    }
    return new String(buf); 
  }
  public boolean equals(DataKeeperJ para) {
    System.out.println("DataKeeperJ: equals not implemented yet");
    return false ;
  }

  /** is this used anywhere? */
  public Object getData() {
    switch(dim) {
    case 1: return data1;
    case 2: return data2; 
    case 3: return data3;
    case 4: return data4;
    case 5: return data5;
    case 6: return data6;
    default: System.err.println("Error: dim not implemented yet"); break;
    }
    return null;
  }

  public void setData(float[]           data) {
    this.data1 = data;
    this.dim   = 1;
  }
  public void setData(float[][]         data) {
    this.data2 = data;
    this.dim   = 2;
  }
  public void setData(float[][][]       data) {
    this.data3 = data;
    this.dim   = 3;
  }
  public void setData(float[][][][]     data) {
    this.data4 = data;
    this.dim   = 4;
  }
  public void setData(float[][][][][]   data) {
    this.data5 = data;
    this.dim   = 5;
  }
  public void setData(float[][][][][][] data) {
    this.data6 = data;
    this.dim   = 6;
  }

  public int getDimension() {
    return dim;
  }
}





