package  sep.operator;

import 	 jam.vector.*;
import 	 jam.operator.*;
import juice.vector.*;
import juice.operator.BlockLinOpAdj;

/**
 * This class implements the Hessian of <TT>SepBiLinearOperator</TT>.
 * To create the Hessian of a Bilinear operator, B, simply invoke: <BR>
 * <TT>new SepBiLinOpHessian(B)</TT>
 *
 * <BR><!--#include file="./source.inc"--><BR>
 * 
 * @author     Joel Schroeder
 */

public class SepBiLinOpHessian extends SepBiLinearOperator {
  SepBiLinearOperator B;

  public SepBiLinOpHessian(SepBiLinearOperator B) {
    super(B.getDomain(), B.getDomain(), B.getRange());
    this.B = B;
  }

  protected void apply(boolean add, Vector in1, Vector in2,
		       Vector  out, int result) {
    // A Hessian is symmetric in its inputs
    if (result == SepBiLinearOperator.kInput2) {
      result = SepBiLinearOperator.kInput1;
      Vector tmp = in1;
      in1 = in2;
      in2 = tmp;
    }

    Vector in11 = ((ProductVector) in1).getSubVector(0);
    Vector in12 = ((ProductVector) in1).getSubVector(1);
    Vector in21 = ((ProductVector) in2).getSubVector(0);
    Vector in22 = ((ProductVector) in2).getSubVector(1);

    if (in1 == in2) { // for efficiency AND correctness
      switch (result) {
      case SepBiLinearOperator.kInput1:
	Vector temp = in11.getSpace().newMember();
	temp.copyFrom(in11);
	B.apply(add , in11, in22, out, SepBiLinearOperator.kInput1);
	B.apply(add , temp, in22, out, SepBiLinearOperator.kInput2);
	break;
      case SepBiLinearOperator.kOutput:
	if (add) {
	  Vector img = out.getSpace().newMember();
	  B.apply(false, in11, in22, img, SepBiLinearOperator.kOutput);
	  out.addScale(2F, img);
	}
	else {
	  B.apply(false, in11, in22, out, SepBiLinearOperator.kOutput);
	  out.scale(2F);
	}
	break;
      }
    }
    else {
      switch (result) {
      case SepBiLinearOperator.kInput1:
	if (in11 != in21) { // this check required for correctness
	  B.apply(add , in11, in22, out, SepBiLinearOperator.kInput1);
	  B.apply(add , in21, in12, out, SepBiLinearOperator.kInput2);
	}
	else {
	  Vector temp = in11.getSpace().newMember();
	  temp.copyFrom(in11);
	  B.apply(add , in11, in22, out, SepBiLinearOperator.kInput1);
	  B.apply(add , temp, in12, out, SepBiLinearOperator.kInput2);
	}
	break;
      case SepBiLinearOperator.kOutput:
	if (in1 != in2) {
	  B.apply(add , in11, in22, out, SepBiLinearOperator.kOutput);
	  B.apply(true, in21, in12, out, SepBiLinearOperator.kOutput);
	}
	else { // this version (in1 == in2) should be faster
	  if (add) {
	    Vector img = out.getSpace().newMember();
	    B.apply(false, in11, in22, img, SepBiLinearOperator.kOutput);
	    out.addScale(2F, img);
	  }
	  else {
	    B.apply(false, in11, in22, out, SepBiLinearOperator.kOutput);
	    out.scale(2F);
	  }
	}
	break;
      }
    }
  }

  public Operator getOp1(Vector d1) {
    return getOp2(d1); // Hessian is symmetric
  }
  public Operator getOp2(Vector d2) {
    return B.getDerivativeAt(d2);
  }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof SepBiLinOpHessian)) return false;
    return this.B.equals(((SepBiLinOpHessian) obj2).B);
  }

  public jam.JamObject copy() {
    System.err.println("SepBiLinOpHessian.copy() not implemented.\n");
    return null;
  }
}
