package  sep.operator;

import 	 jam.vector.*;
import 	 jam.operator.*;
import juice.vector.*;
import juice.operator.BlockLinOpAdj;

/**
 * This class implements a lot of often duplicated code.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * A subclass need only implement apply() and the constructors.<BR>
 * 
 */
public abstract class SepBiLinearOperator
implements BiLinearOperatorHessian {
  public static final int kInput1 = 0;
  public static final int kInput2 = 1;
  public static final int kOutput = 2;

  protected ProductSpc domain = null;
  protected Space      range  = null;

  abstract protected void apply(boolean add, Vector in1, Vector in2,
				Vector out, int result);

  public SepBiLinearOperator(Space in1, Space in2, Space out) {
    domain = new ProductSpc(in1, in2);
    range  = out;
  }

  /** copy constructor */
  protected SepBiLinearOperator(SepBiLinearOperator src) {
    this.domain = src.domain;
    this.range  = src.range;
  }

  protected boolean setSpace(int spc, Space space) {
    switch(spc) {
    case kInput1:
      setDomain1(space);
      break;
    case kInput2:
      setDomain2(space);
      break;
    case kOutput:
      setRange  (space);
      break;
    default:
      return false;
    }
    return true;
  }

  public boolean equals(Object obj2) {
    if (!(obj2 instanceof SepBiLinearOperator )) return false;
    SepBiLinearOperator sbl2 = (SepBiLinearOperator) obj2;
    return (getDomain().equals(sbl2.getDomain()) &&
	    getRange ().equals(sbl2.getRange ()));
  }

  /**
   * <STRONG>WARNING: not yet implemented</STRONG>
   */
  public hasAdjoint getPartialAdjoint(Vector r) {
    System.err.println("Error: SepBiLinearOperator" +
			".getPartialAdjoint() not yet implemented.\n" );
    return null; }

  public Space    getDomain1() { return domain.getSubSpace(0); }
  public Space    getDomain2() { return domain.getSubSpace(1); }
  public Space    getRange  () { return range  ; }

  protected void  setDomain1(Space spc) {
    domain.setSubSpace(0, spc);
  }
  protected void  setDomain2(Space spc) {
    domain.setSubSpace(1, spc);
  }
  protected void  setRange  (Space spc) {
    range = spc;
  }

  public Vector imageBinary   (Vector d1, Vector d2          ) {
    Vector r = range.newMember();
    apply(false, d1, d2, r, kOutput);
    return r;
  }

  public void   imageBinary   (Vector d1, Vector d2, Vector r) {
    apply(false, d1, d2, r, kOutput);
  }

  public void   addImageBinary(Vector d1, Vector d2, Vector r) {
    apply(true , d1, d2, r, kOutput);
  }

  public void   residualBinary(Vector d1, Vector d2,
			       Vector b , Vector r) {
    r.neg(b);
    addImageBinary(d1, d2, r);
  }

  // **************** NONLINEAR OPERATOR FORMS *****************

  public Space getDomain() { return domain; }

  public Vector       image   (Vector d) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in InternalConvolution.image(): " +
			 "Vector argument not a ProductVector");
      return null;
    }
    return imageBinary(((ProductVector) d).getSubVector(0),
		       ((ProductVector) d).getSubVector(1));
  }

  public void         image   (Vector d, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in InternalConvolution.image(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    imageBinary(((ProductVector) d).getSubVector(0),
		((ProductVector) d).getSubVector(1), r);
  }

  public void         addImage(Vector d, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in InternalConvolution.addImage(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    addImageBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), r);

  }

  public void         residual(Vector d, Vector b, Vector r) {
    if (! (d instanceof ProductVector)) {
      System.err.println("Error in InternalConvolution.residual(): " +
			 "Vector argument not a ProductVector");
      return;
    }
    residualBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), b, r);

  }

  public Operator getOp1(Vector d1) {
    return new SepBiLinSubOp(this, d1, SepBiLinearOperator.kInput1);
  }
  public Operator getOp2(Vector d2) {
    return new SepBiLinSubOp(this, d2, SepBiLinearOperator.kInput2);
  }

  public LinearOperator getGradientAt  (Vector x) {
    return getDerivativeAt(x).adjoint();
  }
  public hasAdjoint     getDerivativeAt(Vector x) {
    Vector d1 = ((ProductVector) x).getSubVector(0);
    Vector d2 = ((ProductVector) x).getSubVector(1);

    BlockLinearOperatorAdjoint
      der = new BlockLinOpAdj(getDomain(), getRange());
    der.setOp(0,0, (hasAdjoint) this.getOp2(d2));
    der.setOp(0,1, (hasAdjoint) this.getOp1(d1));

    return der;
  }
  public BiLinearOperatorAdjoint getHessianAt(Vector x) {
    return new SepBiLinOpHessian(this);
  }

  // I/O
  public String toString() {
    return ("Domain1:\n" + getDomain1() + "\n" +
	    "Domain2:\n" + getDomain2() + "\n" +
	    "Range  :\n" + getRange  () );
  }
}

/**
 * <BR><!--#include file="./source.inc"--><BR>
 */
class SepBiLinSubOp extends SepLinearOperator {
  SepBiLinearOperator B;
  Vector 	      di;
  /** = kInput1 or kInput2 */
  int    	      domKnown;

  public SepBiLinSubOp(SepBiLinearOperator B, Vector d, int givenDomain) {
    this.B  	  = B;  
    this.di 	  = d;
    this.domKnown = givenDomain;
    if ((domKnown != SepBiLinearOperator.kInput1) &&
	(domKnown != SepBiLinearOperator.kInput2))
      System.err.println("Error in SepBiLinSubOp constructor: " +
			 "integer givenDomain = " + givenDomain +
			 "\nIt should be " +
			 SepBiLinearOperator.kInput1 + " or " +
			 SepBiLinearOperator.kInput2);
  }

  protected SepBiLinSubOp(SepBiLinSubOp src) {
    super(src);
    this.B        = (SepBiLinearOperator) src.B .copy();
    this.di       = (Vector)              src.di.copy();
    this.domKnown =                       src.domKnown;
  }

  protected void apply(boolean add, Vector d, Vector r) {
    if (domKnown == SepBiLinearOperator.kInput1)
      if (isFwd) 
	B.apply(add, di, d , r, SepBiLinearOperator.kOutput);
      else
	B.apply(add, di, d , r, SepBiLinearOperator.kInput2);
    else
      if (isFwd) 
	B.apply(add, d , di, r, SepBiLinearOperator.kOutput);
      else
	B.apply(add, d , di, r, SepBiLinearOperator.kInput1);
  }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof SepBiLinSubOp)) return false;
    SepBiLinSubOp blo2 = (SepBiLinSubOp) obj2;
    return (this.B .equals(blo2.B ) &&
	    this.di.equals(blo2.di) &&
	    this.domKnown == blo2.domKnown);
  }

  public jam.JamObject copy() { return new SepBiLinSubOp(this); }
}

/**
 * <BR><!--#include file="./source.inc"--><BR>
 */
final class SepBiLinHessian extends SepBiLinearOperator {
  SepBiLinearOperator B;

  public SepBiLinHessian(SepBiLinearOperator B) {
    super(B.getRange(), B.getRange(), B.getDomain());
    this.B = B;
  }

  protected void apply(boolean add, Vector in1, Vector in2,
		       Vector  out, int result) {
    // A Hessian is symmetric in its inputs
    if (result == SepBiLinearOperator.kInput2) {
      result = SepBiLinearOperator.kInput1;
      Vector tmp = in1;
      in1 = in2;
      in2 = tmp;
    }

    Vector in11 = ((ProductVector) in1).getSubVector(0);
    Vector in12 = ((ProductVector) in1).getSubVector(1);
    Vector in21 = ((ProductVector) in2).getSubVector(0);
    Vector in22 = ((ProductVector) in2).getSubVector(1);

    if (in1 == in2) { // for efficiency AND correctness
      switch (result) {
      case SepBiLinearOperator.kInput1:
	Vector temp = in11.getSpace().newMember();
	temp.copyFrom(in11);
	B.apply(add , in11, in22, out, SepBiLinearOperator.kInput1);
	B.apply(add , temp, in22, out, SepBiLinearOperator.kInput2);
	break;
      case SepBiLinearOperator.kOutput:
	if (add) {
	  Vector img = out.getSpace().newMember();
	  B.apply(false, in11, in22, img, SepBiLinearOperator.kOutput);
	  out.addScale(2F, img);
	}
	else {
	  B.apply(false, in11, in22, out, SepBiLinearOperator.kOutput);
	  out.scale(2F);
	}
	break;
      }
    }
    else {
      switch (result) {
      case SepBiLinearOperator.kInput1:
	if (in11 != in21) { // this check required for correctness
	  B.apply(add , in11, in22, out, SepBiLinearOperator.kInput1);
	  B.apply(add , in21, in12, out, SepBiLinearOperator.kInput2);
	}
	else {
	  Vector temp = in11.getSpace().newMember();
	  temp.copyFrom(in11);
	  B.apply(add , in11, in22, out, SepBiLinearOperator.kInput1);
	  B.apply(add , temp, in12, out, SepBiLinearOperator.kInput2);
	}
	break;
      case SepBiLinearOperator.kOutput:
	if (in1 != in2) {
	  B.apply(add , in11, in22, out, SepBiLinearOperator.kOutput);
	  B.apply(true, in21, in12, out, SepBiLinearOperator.kOutput);
	}
	else { // this version (in1 == in2) should be faster
	  if (add) {
	    Vector img = out.getSpace().newMember();
	    B.apply(false, in11, in22, img, SepBiLinearOperator.kOutput);
	    out.addScale(2F, img);
	  }
	  else {
	    B.apply(false, in11, in22, out, SepBiLinearOperator.kOutput);
	    out.scale(2F);
	  }
	}
	break;
      }
    }
  }

  public Operator getOp1(Vector d1) {
    return getOp2(d1); // Hessian is symmetric
  }
  public Operator getOp2(Vector d2) {
    return B.getDerivativeAt(d2);
  }

  public boolean equals(Object obj2) {
    if (this == obj2) return true;
    if (! (obj2 instanceof SepBiLinHessian)) return false;
    return this.B.equals(((SepBiLinHessian) obj2).B);
  }

  public jam.JamObject copy() {
    System.err.println("SepBiLinHessian.copy() not implemented.\n");
    return null;
  }
}
