package  sep.operator;

import 	 jam.vector.*;
import 	 jam.operator.*;
import juice.vector.*;
import juice.operator.BlockLinOpAdj;

/**
 * This class implements a lot of often duplicated code.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * A subclass need only implement apply() and the constructors.<BR>
 *
 * @author      Joel Schroeder
 */
public abstract class SepBinaryOperator
implements BinaryOperator {
  protected ProductSpc domain = null;
  protected Space      range  = null;

  abstract protected void apply(boolean add,
				Vector in1, Vector in2, Vector out);

  public SepBinaryOperator(Space in1, Space in2, Space out) {
    domain = new ProductSpc(in1, in2);
    range  = out;
  }

  /** copy constructor */
  protected SepBinaryOperator(SepBinaryOperator src) {
    this.domain = src.domain;
    this.range  = src.range;
  }

  public boolean equals(Object obj2) {
    if (!(obj2 instanceof SepBinaryOperator)) return false;
    SepBinaryOperator sbo2 = (SepBinaryOperator) obj2;
    return (getDomain().equals(sbo2.getDomain()) &&
	    getRange ().equals(sbo2.getRange ()));
  }

  public Space    getDomain1() { return domain.getSubSpace(0); }
  public Space    getDomain2() { return domain.getSubSpace(1); }
  public Space    getDomain () { return domain; }
  public Space    getRange  () { return range ; }

  public Vector imageBinary   (Vector d1, Vector d2          ) {
    Vector r = range.newMember();
    apply(false, d1, d2, r);
    return r;
  }

  public void   imageBinary   (Vector d1, Vector d2, Vector r) {
    apply(false, d1, d2, r);
  }

  public void   addImageBinary(Vector d1, Vector d2, Vector r) {
    apply(true , d1, d2, r);
  }

  public void   residualBinary(Vector d1, Vector d2,
			       Vector b , Vector r) {
    r.neg(b);
    addImageBinary(d1, d2, r);
  }

  public Vector       image   (Vector d) {
    return imageBinary(((ProductVector) d).getSubVector(0),
		       ((ProductVector) d).getSubVector(1));
  }

  public void         image   (Vector d, Vector r) {
    imageBinary(((ProductVector) d).getSubVector(0),
		((ProductVector) d).getSubVector(1), r);
  }

  public void         addImage(Vector d, Vector r) {
    addImageBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), r);

  }

  public void         residual(Vector d, Vector b, Vector r) {
    residualBinary(((ProductVector) d).getSubVector(0),
		   ((ProductVector) d).getSubVector(1), b, r);

  }

  public Operator getOp1(Vector d1) {
    return new SepSubBinaryOp(d1, 1, this);
  }

  public Operator getOp2(Vector d2) {
    return new SepSubBinaryOp(d2, 2, this);
  }

  // I/O
  public String toString() {
    return ("Domain1:\n" + getDomain1() + "\n" +
	    "Domain2:\n" + getDomain2() + "\n" +
	    "Range  :\n" + getRange  () );
  }
}

final class SepSubBinaryOp extends SepOperator {
  SepBinaryOperator B;
  Vector            d;
  int               domGiven;

  /** domGiven should be 1 or 2 */
  public SepSubBinaryOp(Vector d, int domGiven, SepBinaryOperator B) {
    super(d.getSpace(), B.getRange());
    this.B        = B;
    this.d        = d;
    this.domGiven = domGiven;
  }

  protected void apply(boolean add, Vector dom, Vector ran) {
    if (domGiven == 1) {
      if (add)
	B.addImageBinary(d, dom, ran);
      else
	B.   imageBinary(d, dom, ran);
    }
    else {
      if (add)
	B.addImageBinary(dom, d, ran);
      else
	B.   imageBinary(dom, d, ran);
    }
  }

  public jam.JamObject copy() {
    System.err.println("SepBinaryOperator.copy() not implemented.\n");
    return null;
  }
}
