package  sep.operator;

import 	 jam.vector.*;
import 	 jam.operator.*;

/**
 * This class implements a lot of often duplicated code.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * A subclass need only implement apply() and the constructors.
 */
public abstract class SepOperator implements Operator
{
  /** domain of the operator */
  protected Space domain;
  /** range  of the operator */
  protected Space range;

  public SepOperator() {}

  public SepOperator(Space domain, Space range) {
    setDomain(domain);
    setRange (range );
  }

  public SepOperator(Space domainAndRange) {
    this(domainAndRange, domainAndRange);
  }

  /** copy constructor (shallow copy) */
  public SepOperator(SepOperator src) {
    this.domain = src.domain;
    this.range  = src.range ;
  }

  /**
   * Apply the operator:<BR><BR>
   * ran = this(dom)
   *
   * @param add if true, perform ran += this(dom) instead of ran = this(dom)
   * @param in  Vector in the domain (will NOT be altered)
   * @param out Vector in the range  (will be altered)
   */
  protected abstract void  apply     (boolean add, Vector dom, Vector ran);

  protected void  setDomain (Space domain) { this.domain = domain; }
  protected void  setRange  (Space range ) { this.range  = range ; }

  // ----------------------------------------------------------------------
  // ---------- The remaining methods are required by the superclass ------
  // ----------------------------------------------------------------------

  public boolean equals(Object obj) {
    if (! (obj instanceof SepOperator)) return false;
    SepOperator op2 = (SepOperator) obj;
    return (this.domain.equals(op2.domain) &&
	    this.range .equals(op2.range ));
  }

  public Space   getDomain () { return domain; }
  public Space   getRange  () { return range ; }

  public Vector  image(Vector d) {
    Vector r = getRange().newMember();
    image(d, r);
    return r;
  }

  public void    image(Vector d, Vector r) { apply(false, d, r); }
  public void addImage(Vector d, Vector r) { apply(true , d, r); }

  public void residual(Vector d, Vector b, Vector r) {
    r.neg(b);
    addImage(d, r);
  }
}
