package  sep.operator;

import 	 jam.vector.*;
import 	 jam.operator.*;
import juice.operator.ZeroBiLinOp;

/**
 * This class implements a lot of often duplicated code.<BR>
 * <BR><!--#include file="./source.inc"--><BR>
 * A subclass need only implement apply() and the constructors.
 * 
 */
public abstract class SepSelfAdjointOperator
implements hasAdjoint, hasHessian
{
  /** the common domain and range of the operator */
  protected Space         space;

  public SepSelfAdjointOperator() {}

  public SepSelfAdjointOperator(Space domainAndRange) {
    setSpace(domainAndRange);
  }

  /** copy constructor (shallow copy) */
  public SepSelfAdjointOperator(SepSelfAdjointOperator src) {
    this.space = src.space;
  }

  /**
   * Apply the operator:<BR><BR>
   * Images ran = A(dom), where A indicates the operator, this.
   *
   * <BR>
   * CONTRAST THIS INTERFACE WITH THAT OF SepLinearOperator.apply()
   * <BR>
   *
   * @param add Performs ran += A(dom) instead of ran = A(dom) if true
   * @param dom Vector from which to map (remains unaltered)
   * @param ran Vector onto which to map (becomes altered)
   */
  protected abstract  void  apply     (boolean add, Vector dom, Vector ran);

  protected void  setSpace  (Space space) { this.space = space; }

  // ----------------------------------------------------------------------
  // ---------- The remaining methods are required by the superclass ------
  // ----------------------------------------------------------------------

  public boolean equals(Object obj) {
    if (! (obj instanceof SepSelfAdjointOperator)) return false;
    SepSelfAdjointOperator op2 = (SepSelfAdjointOperator) obj;
    return (this.space.equals(op2.space));
  }

  public Space   getDomain () { return space; }
  public Space   getRange  () { return space; }

  public Vector  image(Vector d) {
    Vector r = getRange().newMember();
    image(d, r);
    return r; }

  public void    image(Vector d, Vector r) { apply(false, d, r); }
  public void addImage(Vector d, Vector r) { apply(true , d, r); }

  public void residual(Vector d, Vector b, Vector r) {
    r.neg(b);
    addImage(d, r); }

  public hasAdjoint     adjoint()                 { return this; }
  public LinearOperator getGradientAt  (Vector x) { return this; }
  public hasAdjoint     getDerivativeAt(Vector x) { return this; }

  public BiLinearOperatorAdjoint getHessianAt(Vector x) {
    return new ZeroBiLinOp(this.getDomain(), this.getDomain(),
			   this.getRange());
  }
}
